{-# LANGUAGE EmptyDataDecls #-}

module OkCrypt.Bindings
  ( init,
    randomBytes,
    Material,
    toByteString,
    unsafeMaterialFromByteString,

    -- * Authenticated Encryption with Secret Key
    generateSecretKey,
    toSecretKey,
    generateNonce,
    toPlaintext,
    encryptSecret,
    toCiphertext,
    decryptSecret,

    -- * Key Derivation
    generateSalt,
    deriveSecretKey,

    -- * Generic Hashing
    genericHash,
    genericHashHex,
  )
where

import qualified Data.ByteString as BS
import Foreign.Marshal.Alloc (allocaBytes)
import Foreign.Ptr (nullPtr)
import qualified OkCrypt.Cdefs as C
import OkCrypt.Exception
import OkStd

-- | Initializes Libsodium. Throws `ExceptionInitializationFailed` if initialization fails.
init :: IO ()
init =
  C.sodium_init >>= \case
    0 -> pure ()
    1 -> pure ()
    _ -> throwM ExceptionInitializationFailed

-- | Returns requested number of random bytes
randomBytes :: Int -> IO ByteString
randomBytes size = allocaBytes size $ \ptr -> do
  C.randombytes_buf ptr (fromIntegral size)
  BS.packCStringLen (ptr, size)

-- | Material used in cryptographic operations: secret and public keys,
-- plaintext, ciphertext, signature, etc.
newtype Material _role = Material ByteString

data SecretKey

data Nonce

data Plaintext

data Ciphertext

data Salt

-- | Converts `Material` to `ByteString`
toByteString :: Material any -> ByteString
toByteString (Material s) = s

-- | UNSAFE. Converts any `ByteString` to `Material` without asking questions.
unsafeMaterialFromByteString :: ByteString -> Material any
unsafeMaterialFromByteString = Material

-- * Authenticated Encryption with Secret Key

-- | Generates new random secret key
generateSecretKey :: IO (Material SecretKey)
generateSecretKey = randomMaterial $ fromIntegral C.crypto_secretbox_KEYBYTES

-- | Wraps `ByteString` into secret key. Throws `ExceptionInvalidLength` if given data is not 32 bytes long.
toSecretKey :: MonadThrow m => ByteString -> m (Material SecretKey)
toSecretKey sk
  | BS.length sk == fromIntegral C.crypto_secretbox_KEYBYTES =
    pure $ Material sk
  | otherwise =
    throwM ExceptionInvalidLength

-- | Generates new nonce
generateNonce :: IO (Material Nonce)
generateNonce = randomMaterial $ fromIntegral C.crypto_secretbox_NONCEBYTES

-- | Wraps `ByteString` into plaintext
toPlaintext :: ByteString -> Material Plaintext
toPlaintext = Material

-- | Encrypts plaintext using specified secret key and nonce
encryptSecret :: Material Plaintext -> Material SecretKey -> Material Nonce -> IO (Material Ciphertext)
encryptSecret (Material p) (Material k) (Material n) =
  BS.useAsCStringLen p $ \(pPtr, pSize) ->
    BS.useAsCString k $ \kPtr ->
      BS.useAsCString n $ \nPtr ->
        let cSize = pSize + fromIntegral C.crypto_secretbox_MACBYTES
         in allocaBytes cSize $ \cPtr -> do
              code <- C.crypto_secretbox_easy cPtr pPtr (fromIntegral pSize) nPtr kPtr
              if code == C.sodiumSuccess
                then Material <$> BS.packCStringLen (cPtr, cSize)
                else throwM $ ExceptionLibsodium (fromIntegral code)

-- | Wraps `ByteString` into ciphertext. Throws `ExceptionInvalidLength` if given data is not at least 16 bytes long
toCiphertext :: MonadThrow m => ByteString -> m (Material Ciphertext)
toCiphertext ct
  | BS.length ct >= fromIntegral C.crypto_secretbox_MACBYTES =
    pure $ Material ct
  | otherwise =
    throwM ExceptionInvalidLength

-- | Decrypts ciphertext using specified secret key and nonce
decryptSecret :: Material Ciphertext -> Material SecretKey -> Material Nonce -> IO (Material Plaintext)
decryptSecret (Material c) (Material k) (Material n) =
  BS.useAsCStringLen c $ \(cPtr, cSize) ->
    BS.useAsCString k $ \kPtr ->
      BS.useAsCString n $ \nPtr ->
        let pSize = cSize - fromIntegral C.crypto_secretbox_MACBYTES
         in allocaBytes pSize $ \pPtr -> do
              code <- C.crypto_secretbox_open_easy pPtr cPtr (fromIntegral cSize) nPtr kPtr
              if code == C.sodiumSuccess
                then Material <$> BS.packCStringLen (pPtr, pSize)
                else throwM $ ExceptionLibsodium (fromIntegral code)

-- * Key Derivation

-- | Generates new salt
generateSalt :: IO (Material Salt)
generateSalt = randomMaterial $ fromIntegral C.crypto_pwhash_SALTBYTES

-- | Derives secret key from given password and salt using Argon2ID13 algorithm
-- and @INTERACTIVE@ limits for operations and memory
deriveSecretKey :: ByteString -> Material Salt -> IO (Material SecretKey)
deriveSecretKey p (Material s) =
  BS.useAsCStringLen p $ \(pPtr, pSize) ->
    BS.useAsCString s $ \sPtr ->
      let kSize = fromIntegral C.crypto_secretbox_KEYBYTES
       in allocaBytes kSize $ \kPtr -> do
            code <-
              C.crypto_pwhash
                kPtr
                (fromIntegral kSize)
                pPtr
                (fromIntegral pSize)
                sPtr
                C.crypto_pwhash_OPSLIMIT_INTERACTIVE
                C.crypto_pwhash_MEMLIMIT_INTERACTIVE
                C.crypto_pwhash_ALG_ARGON2ID13
            if code == C.sodiumSuccess
              then Material <$> BS.packCStringLen (kPtr, kSize)
              else throwM $ ExceptionLibsodium (fromIntegral code)

-- * Generic Hashing

-- | Returns fingerprint of the specified input data in binary form
genericHash :: ByteString -> IO ByteString
genericHash input =
  BS.useAsCStringLen input $ \(inPtr, inSize) ->
    let outSize = fromIntegral C.crypto_generichash_BYTES
     in allocaBytes outSize $ \outPtr -> do
          code <- C.crypto_generichash outPtr (fromIntegral outSize) inPtr (fromIntegral inSize) nullPtr 0
          if code == C.sodiumSuccess
            then BS.packCStringLen (outPtr, outSize)
            else throwM $ ExceptionLibsodium (fromIntegral code)

-- | Returns fingerprint of the specified input data as hexadecimal string
genericHashHex :: ByteString -> IO Text
genericHashHex input =
  toHexString <$> genericHash input

randomMaterial :: Int -> IO (Material r)
randomMaterial s =
  Material <$> randomBytes s
