module OkSqlite.Bindings
  ( Database,
    OpenMode (..),
    Statement,
    open,
    close,
    prepare,
    finalize,
    step,
    columnCount,
    columnInt64,
    columnDouble,
    columnText,
    columnBlob,
    columnIsNull,
    bindInt64,
    bindDouble,
    bindText,
    bindBlob,
    bindNull,
  )
where

import Control.Monad.Catch (catchAll)
import Data.ByteString (packCString, packCStringLen, useAsCString, useAsCStringLen)
import Data.IORef (IORef, modifyIORef', newIORef, readIORef)
import Foreign (Ptr, alloca, nullPtr, peek)
import Foreign.C.Types (CInt)
import qualified OkSqlite.Cdefs as C
import OkSqlite.Exception
import OkStd

-- | SQLite database handle
newtype Database = Database (Ptr C.Csqlite3)
  deriving (Eq, Show)

-- | Database opening mode
data OpenMode
  = -- | Open database in read-only mode; raise an error if it does not exist
    OpenReadOnly
  | -- | Open database in read-write mode; raise an error if it does not exist
    OpenReadWrite
  | -- | Open database in read-write mode; create a new one if it does not exist
    OpenReadWriteCreate
  deriving (Show, Eq)

-- | SQLite statement handle
newtype Statement = Statement (Ptr C.Csqlite3_stmt, IORef Bool)
  deriving (Eq)

-- | Opens SQLite database file. Special value @:memory:@ creates new in-memory
-- database.
open :: Text -> OpenMode -> IO Database
open filename openMode =
  useAsCString (encodeUtf8 filename) $ \filenameCstr ->
    alloca $ \dbPtr -> do
      status <- C.sqlite3_open_v2 filenameCstr dbPtr flags nullPtr
      db <- peek dbPtr
      okOrError (Database db) db status `catchAll` \exc -> do
        _ <- C.sqlite3_close_v2 db
        throwM exc
  where
    flags = case openMode of
      OpenReadOnly -> C.c_SQLITE_OPEN_READONLY
      OpenReadWrite -> C.c_SQLITE_OPEN_READWRITE
      OpenReadWriteCreate -> C.c_SQLITE_OPEN_READWRITE .|. C.c_SQLITE_OPEN_CREATE

-- | Closes SQLite database
close :: Database -> IO ()
close (Database db) =
  C.sqlite3_close_v2 db >>= okOrError () db

-- | Creates new SQLite statement
prepare :: Database -> Text -> IO Statement
prepare (Database db) sql =
  useAsCString (encodeUtf8 sql) $ \sqlCstr ->
    alloca $ \stmtPtr -> do
      status <- C.sqlite3_prepare_v2 db sqlCstr (-1) stmtPtr nullPtr
      stmt <- peek stmtPtr
      isFinalizedRef <- newIORef False
      okOrError (Statement (stmt, isFinalizedRef)) db status

-- | Destroys SQLite statement
finalize :: Statement -> IO ()
finalize (Statement (stmt, isFinalizedRef)) =
  readIORef isFinalizedRef >>= \case
    True -> pure ()
    False -> do
      db <- C.sqlite3_db_handle stmt
      modifyIORef' isFinalizedRef (const True)
      C.sqlite3_finalize stmt >>= okOrError () db

-- | Evaluates the SQLite statement. Returns True when after evaluation a new
-- row of output is available. Returns False when all the output rows were
-- already returned.
step :: Statement -> IO Bool
step (Statement (stmt, isFinalized)) = do
  checkIsFinalized isFinalized
  db <- C.sqlite3_db_handle stmt
  C.sqlite3_step stmt >>= processStatusCode db
  where
    processStatusCode db status
      | status == C.c_SQLITE_ROW =
        pure True
      | status == C.c_SQLITE_DONE =
        pure False
      | otherwise =
        mkError db status >>= throwM

-- | Returns number of columns in the result set of the statement
columnCount :: Statement -> IO Int
columnCount (Statement (stmt, isFinalized)) = do
  checkIsFinalized isFinalized
  fromIntegral <$> C.sqlite3_column_count stmt

-- | Reads Int64 value from the specified column
columnInt64 :: Statement -> Int -> IO Int64
columnInt64 (Statement (stmt, isFinalized)) columnIndex = do
  checkIsFinalized isFinalized
  fromIntegral <$> C.sqlite3_column_int64 stmt (fromIntegral columnIndex)

-- | Reads Double value from the specified column
columnDouble :: Statement -> Int -> IO Double
columnDouble (Statement (stmt, isFinalized)) columnIndex = do
  checkIsFinalized isFinalized
  realToFrac <$> C.sqlite3_column_double stmt (fromIntegral columnIndex)

-- | Reads Text value from the specified column
columnText :: Statement -> Int -> IO Text
columnText (Statement (stmt, isFinalized)) columnIndex = do
  checkIsFinalized isFinalized
  cstr <- C.sqlite3_column_text stmt (fromIntegral columnIndex)
  decodeUtf8 <$> packCString cstr

-- | Reads ByteString value from the specified column
columnBlob :: Statement -> Int -> IO ByteString
columnBlob (Statement (stmt, isFinalized)) columnIndex = do
  checkIsFinalized isFinalized
  len <- C.sqlite3_column_bytes stmt (fromIntegral columnIndex)
  cstr <- C.sqlite3_column_blob stmt (fromIntegral columnIndex)
  packCStringLen (cstr, fromIntegral len)

-- | Checks if the column value is null
columnIsNull :: Statement -> Int -> IO Bool
columnIsNull (Statement (stmt, isFinalized)) columnIndex = do
  checkIsFinalized isFinalized
  (== C.c_SQLITE_NULL) <$> C.sqlite3_column_type stmt (fromIntegral columnIndex)

-- | Binds Int64 value to the specified named parameter
bindInt64 :: Statement -> Text -> Int64 -> IO ()
bindInt64 (Statement (stmt, isFinalized)) paramName value = do
  checkIsFinalized isFinalized
  columnIndex <- findParam stmt paramName
  status <- C.sqlite3_bind_int64 stmt columnIndex (fromIntegral value)
  db <- C.sqlite3_db_handle stmt
  okOrError () db status

-- | Binds Double value to the specified named parameter
bindDouble :: Statement -> Text -> Double -> IO ()
bindDouble (Statement (stmt, isFinalized)) paramName value = do
  checkIsFinalized isFinalized
  columnIndex <- findParam stmt paramName
  status <- C.sqlite3_bind_double stmt columnIndex (realToFrac value)
  db <- C.sqlite3_db_handle stmt
  okOrError () db status

-- | Binds Text value to the specified named parameter
bindText :: Statement -> Text -> Text -> IO ()
bindText (Statement (stmt, isFinalized)) paramName value = do
  checkIsFinalized isFinalized
  columnIndex <- findParam stmt paramName
  useAsCString (encodeUtf8 value) $ \valueCstr -> do
    status <- C.sqlite3_bind_text stmt columnIndex valueCstr (-1) C.c_SQLITE_TRANSIENT
    db <- C.sqlite3_db_handle stmt
    okOrError () db status

-- | Binds ByteString value to the specified named parameter
bindBlob :: Statement -> Text -> ByteString -> IO ()
bindBlob (Statement (stmt, isFinalized)) paramName value = do
  checkIsFinalized isFinalized
  columnIndex <- findParam stmt paramName
  useAsCStringLen value $ \(valueCstr, len) -> do
    status <- C.sqlite3_bind_blob stmt columnIndex valueCstr (fromIntegral len) C.c_SQLITE_TRANSIENT
    db <- C.sqlite3_db_handle stmt
    okOrError () db status

-- | Binds null value to the specified named parameter
bindNull :: Statement -> Text -> IO ()
bindNull (Statement (stmt, isFinalized)) paramName = do
  checkIsFinalized isFinalized
  columnIndex <- findParam stmt paramName
  status <- C.sqlite3_bind_null stmt columnIndex
  db <- C.sqlite3_db_handle stmt
  okOrError () db status

checkIsFinalized :: IORef Bool -> IO ()
checkIsFinalized isFinalizedRef =
  readIORef isFinalizedRef >>= \case
    True -> throwM ExceptionStatementIsFinalized
    False -> pure ()

findParam :: Ptr C.Csqlite3_stmt -> Text -> IO CInt
findParam stmt paramName =
  useAsCString (encodeUtf8 paramName) (C.sqlite3_bind_parameter_index stmt) >>= \case
    0 -> throwM $ ExceptionParameterNotFound paramName
    columnIndex -> pure columnIndex

okOrError :: a -> Ptr C.Csqlite3 -> C.Cerr -> IO a
okOrError val db status
  | db == nullPtr =
    throwM ExceptionDatabaseHandleIsNull
  | status == C.c_SQLITE_OK =
    pure val
  | otherwise =
    mkError db status >>= throwM

mkError :: Ptr C.Csqlite3 -> C.Cerr -> IO ExceptionOkSqlite
mkError db status = do
  errmsgCstr <- C.sqlite3_errmsg db
  errmsg <- decodeUtf8 <$> packCString errmsgCstr
  pure $ ExceptionSQLite (fromIntegral status) errmsg
