{-# LANGUAGE ExistentialQuantification #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

module OkSqlite.Dsl
  ( -- * Database
    Database,
    OpenMode (..),
    open,
    close,
    withDatabase,
    inTransaction,

    -- * Statements
    Statement (..),
    getCode,
    exec,

    -- * Bindings
    Bind (..),
    noBind,

    -- * Results
    Result,
    noResult,
    singleRow,
    rowMaybe,
    rowVector,
    Row,
    column,
  )
where

import Control.Monad.Catch (catchAll)
import Data.IORef (IORef, modifyIORef', newIORef, readIORef)
import qualified Data.Vector as V
import OkSqlite.Bindings (Database, OpenMode, close, open)
import qualified OkSqlite.Bindings as B
import OkSqlite.Exception
import OkSqlite.Values
import OkStd

-- | @withDatabase dbPath openMode fn@ opens new database connection to the
-- database under specified @dbPath@, runs @fn@ on it, and closes the
-- connection regardless of @fn@ raised an exception or not. All exceptions
-- from @fn@ are re-raised after closing the connection.
withDatabase :: Text -> OpenMode -> (Database -> IO a) -> IO a
withDatabase filename openMode =
  bracket (open filename openMode) close

-- | @inTransaction db isImmediate fn@ runs given function @fn@ inside SQLite
-- transaction. It commits changes after succesful execution of @fn@ and tries
-- to rollback when @fn@ throws an exception.
--
-- Setting @isImmediate@ to @True@ causes the database to start a new write
-- transaction immediately, without waiting for a write statement.
--
-- See also: https://www.sqlite.org/lang_transaction.html
inTransaction :: Database -> Bool -> (Database -> IO a) -> IO a
inTransaction db isImmediate fn = do
  exec db beginStmt ()
  result <-
    fn db `catchAll` \exc -> do
      exec db rollbackStmt ()
      throwM exc
  exec db commitStmt ()
  pure result
  where
    beginStmt = Statement (if isImmediate then "BEGIN IMMEDIATE TRANSACTION;" else "BEGIN DEFERRED TRANSACTION;") noBind noResult
    commitStmt = Statement "COMMIT TRANSACTION;" noBind noResult
    rollbackStmt = Statement "ROLLBACK TRANSACTION;" noBind noResult

-- | @Statement sql bindFn resultM@ is a complete definition of an SQLite
-- statement
data Statement binds result = Statement Text (binds -> [Bind]) (Result result)

-- | Returns statement SQL code
getCode :: Statement a b -> Text
getCode (Statement code _ _) = code

-- | Executes SQLite statement
exec :: Database -> Statement binds result -> binds -> IO result
exec db (Statement sql bindsFn resultM) bindValues = do
  stmt <- B.prepare db sql
  bindAndExec stmt `catchAll` \exc -> do
    B.finalize stmt
    throwM exc
  where
    bindAndExec stmt = do
      applyBinds stmt $ bindsFn bindValues
      retrieveResults resultM stmt
    applyBinds _ [] =
      pure ()
    applyBinds stmt (Bind paramName value : bs) = do
      bind stmt paramName value
      applyBinds stmt bs

-- | @Bind paramName value@ describes a bind of @value@ to the specified
-- SQLite statement parameter under @paramName@.
data Bind = forall a. BindableValue a => Bind Text a

-- | Apply no bindings to the statement
noBind :: () -> [Bind]
noBind () = []

-- | @Result@ decodes result(s) of a query
newtype Result a = Result (ResultPriv a)
  deriving (Functor)

-- | Retrieve no rows from the query result set. Useful for `CREATE`, `INSERT`,
-- `UPDATE` or `DELETE` statements.
noResult :: Result ()
noResult = Result noResultPriv

-- | Retrieves single row from the query result set. Raises an error when no
-- rows found.
singleRow :: Row a -> Result a
singleRow = Result . singleRowPriv

-- | Retrieves single row from the query result set. Returns @Nothing@ when no
-- rows found.
rowMaybe :: Row a -> Result (Maybe a)
rowMaybe = Result . rowMaybePriv

-- | Retrieves all rows from the query result set
rowVector :: Row a -> Result (Vector a)
rowVector = Result . rowVectorPriv

-- | @Row@ decodes single result row of a query
newtype Row a = Row (RowPriv a)
  deriving (Functor, Applicative, Monad)

-- | Lifts an individual column extractor to a composable row extractor
column :: ExtractableValue a => Row a
column = Row $ do
  (colRef, stmt) <- ask
  liftIO $ do
    colIdx <- readIORef colRef
    modifyIORef' colRef succ
    extract stmt colIdx

type RowEnv = (IORef Int, B.Statement)

newtype RowPriv a = RowPriv (ReaderT RowEnv IO a)
  deriving (Functor, Applicative, Monad, MonadIO, MonadReader RowEnv, MonadThrow, MonadCatch)

decodeRow :: Row a -> B.Statement -> IO a
decodeRow (Row (RowPriv rowM)) stmt = do
  colRef <- newIORef 0
  runReaderT rowM (colRef, stmt)

newtype ResultPriv a = ResultPriv (ReaderT B.Statement IO a)
  deriving (Functor, Applicative, Monad, MonadIO, MonadReader B.Statement, MonadThrow, MonadCatch)

retrieveResults :: Result a -> B.Statement -> IO a
retrieveResults (Result (ResultPriv resultM)) =
  runReaderT resultM

noResultPriv :: ResultPriv ()
noResultPriv =
  ask >>= \stmt -> liftIO $ do
    _ <- B.step stmt
    B.finalize stmt

singleRowPriv :: Row a -> ResultPriv a
singleRowPriv row =
  rowMaybePriv row >>= \case
    Just r ->
      pure r
    Nothing -> do
      throwM ExceptionNoRowsReturned

rowMaybePriv :: Row a -> ResultPriv (Maybe a)
rowMaybePriv row =
  (V.!? 0) <$!> rowVectorPriv row

rowVectorPriv :: Row a -> ResultPriv (Vector a)
rowVectorPriv row = ask >>= \stmt -> liftIO $ readRows stmt V.empty
  where
    readRows stmt acc =
      B.step stmt >>= \case
        True -> do
          r <- decodeRow row stmt
          readRows stmt (V.snoc acc r)
        False -> do
          B.finalize stmt
          pure acc
