module OkSqlite.Migration
  ( Migration,
    migrate,
    getCurrentVersion,
  )
where

import Data.Vector ((!))
import OkSqlite.Dsl
import OkSqlite.Exception
import OkStd

-- | Migration is a statment without binds that returns no result
type Migration = Statement () ()

-- | Applies migrations to the database in the order they were given
migrate :: Database -> Vector Migration -> IO ()
migrate db' migrations = do
  ensureSchemaTableExists db'
  inTransaction db' True $ apply 0
  where
    apply idx db
      | idx == length migrations =
        pure ()
      | otherwise = do
        let migration = migrations ! idx
            migrationCode = getCode migration
            version = idx + 1
        maybeDBCode <- exec db getMigrationCodeStmt version
        case maybeDBCode of
          Nothing -> do
            exec db migration ()
            exec db insertMigrationStmt (version, migrationCode)
            apply (idx + 1) db
          Just dbCode
            | dbCode == migrationCode ->
              apply (idx + 1) db
          Just dbCode ->
            throwM $ ExceptionMigrationCodeMismatch migrationCode dbCode

-- | Returns the number of migrations applied
getCurrentVersion :: Database -> IO Int64
getCurrentVersion db = do
  schemaTableExists <- exec db schemaTableExistsStmt ()
  fromMaybe 0
    <$> if schemaTableExists
      then exec db countMigrationsStmt ()
      else pure Nothing

ensureSchemaTableExists :: Database -> IO ()
ensureSchemaTableExists db = do
  schemaTableExists <- exec db schemaTableExistsStmt ()
  if schemaTableExists
    then pure ()
    else exec db createSchemaTableStmt ()

createSchemaTableStmt :: Statement () ()
createSchemaTableStmt =
  Statement "CREATE TABLE IF NOT EXISTS dbschema (version INTEGER PRIMARY KEY, code TEXT);" noBind noResult

schemaTableExistsStmt :: Statement () Bool
schemaTableExistsStmt =
  Statement "SELECT COUNT(*) FROM sqlite_schema WHERE type = 'table' AND name = 'dbschema';" noBind ((== 1) <$> singleRow (column :: Row Int64))

insertMigrationStmt :: Statement (Int, Text) ()
insertMigrationStmt =
  Statement "INSERT INTO dbschema (version, code) VALUES (:version, :code);" binds noResult
  where
    binds (version, code) = [Bind ":version" (fromIntegral version :: Int64), Bind ":code" code]

getMigrationCodeStmt :: Statement Int (Maybe Text)
getMigrationCodeStmt =
  Statement "SELECT code FROM dbschema WHERE version = :version;" binds result
  where
    binds version = [Bind ":version" (fromIntegral version :: Int64)]
    result = rowMaybe (column :: Row Text)

countMigrationsStmt :: Statement () (Maybe Int64)
countMigrationsStmt =
  Statement "SELECT version FROM dbschema ORDER BY version DESC LIMIT 1;" noBind (rowMaybe (column :: Row Int64))
