module OkSqlite.Values where

import Data.String (String)
import Data.Text (pack, unpack)
import Data.Time (UTCTime)
import Data.Time.Format (defaultTimeLocale, formatTime, parseTimeM)
import qualified OkSqlite.Bindings as B
import OkStd

-- | Values that could be bound to an SQLite statement parameters
class Show a => BindableValue a where
  bind :: B.Statement -> Text -> a -> IO ()

instance BindableValue Int64 where
  bind = B.bindInt64

instance BindableValue Double where
  bind = B.bindDouble

instance BindableValue Text where
  bind = B.bindText

instance BindableValue ByteString where
  bind = B.bindBlob

instance BindableValue UTCTime where
  bind s f = B.bindText s f . utcTimeToSqliteTimestamp
    where
      utcTimeToSqliteTimestamp = pack . formatTime defaultTimeLocale sqliteTimeFormat

instance BindableValue a => BindableValue (Maybe a) where
  bind stmt paramName Nothing = B.bindNull stmt paramName
  bind stmt paramName (Just value) = bind stmt paramName value

-- | Values that could be extracted from SQLite query results
class Show a => ExtractableValue a where
  extract :: B.Statement -> Int -> IO a

instance ExtractableValue Int64 where
  extract = B.columnInt64

instance ExtractableValue Double where
  extract = B.columnDouble

instance ExtractableValue Text where
  extract = B.columnText

instance ExtractableValue ByteString where
  extract = B.columnBlob

-- | IMPORTANT: this implementation is _partial_. It will fail if timestamp is
-- not in the default SQLite format.
instance ExtractableValue UTCTime where
  extract s c = sqliteTimestampToUTCTime <$!> B.columnText s c
    where
      sqliteTimestampToUTCTime t =
        let st = unpack t
         in case parseTimeM False defaultTimeLocale sqliteTimeFormat st of
              Just ts -> ts
              Nothing -> error $ "extract: can't parse UTCTime from '" <> t <> "'"

instance ExtractableValue a => ExtractableValue (Maybe a) where
  extract stmt columnIndex = do
    isNull <- B.columnIsNull stmt columnIndex
    if isNull then pure Nothing else Just <$> extract stmt columnIndex

sqliteTimeFormat :: String
sqliteTimeFormat = "%0Y-%m-%d %H:%M:%S"
