module TestBindings where

import OkSqlite.Bindings
import OkSqlite.Exception
import OkStd
import Test.HUnit

tests =
  [ "should open and close in-memory database without errors" ~: do
      db <- open ":memory:" OpenReadWrite
      close db,
    "should report an error when opening non-existent DB read-only" ~: do
      Left (ExceptionSQLite code msg) <- try $ open "/this/database/should/not/exist/" OpenReadOnly
      assertEqual "error code" 14 code
      assertEqual "error message" "unable to open database file" msg,
    "should create, evaluate, and finalize valid SQLite statements" ~: do
      db <- open ":memory:" OpenReadWrite
      stmt <- prepare db "CREATE TABLE test (a INTEGER, b REAL, c TEXT, d BLOB, e BLOB);"
      False <- step stmt
      finalize stmt
      stmt <-
        prepare
          db
          "\
          \ INSERT INTO test (a, b, c, d, e) VALUES \
          \ (1, 1.01, 'one', x'0100f1', NULL), \
          \ (2, 2.02, 'に', x'0200f2', 'test'), \
          \ (3, 3.03, 'три', x'0300f3', NULL) \
          \ ;"
      False <- step stmt
      finalize stmt
      stmt <- prepare db "SELECT * FROM test;"
      True <- step stmt
      colCount <- columnCount stmt
      assertEqual "column count" 5 colCount
      val <- columnInt64 stmt 0
      assertEqual "first integer value" 1 val
      val <- columnDouble stmt 1
      assertEqual "first double value" 1.01 val
      val <- columnText stmt 2
      assertEqual "first text value" "one" val
      val <- columnBlob stmt 3
      assertEqual "first blob value" "\x01\x00\xf1" val
      True <- columnIsNull stmt 4
      val <- columnBlob stmt 4
      assertEqual "first nullable blob value" "" val
      True <- step stmt
      val <- columnInt64 stmt 0
      assertEqual "second integer value" 2 val
      val <- columnDouble stmt 1
      assertEqual "second double value" 2.02 val
      val <- columnText stmt 2
      assertEqual "second text value" "に" val
      val <- columnBlob stmt 3
      assertEqual "second blob value" "\x02\x00\xf2" val
      val <- columnBlob stmt 4
      False <- columnIsNull stmt 4
      assertEqual "second nullable blob value" "test" val
      True <- step stmt
      val <- columnInt64 stmt 0
      assertEqual "third integer value" 3 val
      val <- columnDouble stmt 1
      assertEqual "third double value" 3.03 val
      val <- columnText stmt 2
      assertEqual "third text value" "три" val
      val <- columnBlob stmt 3
      assertEqual "third blob value" "\x03\x00\xf3" val
      True <- columnIsNull stmt 4
      val <- columnBlob stmt 4
      assertEqual "third nullable blob value" "" val
      False <- step stmt
      finalize stmt
      close db,
    "should bind parameters" ~: do
      db <- open ":memory:" OpenReadWrite
      stmt <- prepare db "CREATE TABLE test (a INTEGER, b REAL, c TEXT, d BLOB, e BLOB);"
      False <- step stmt
      finalize stmt
      stmt <- prepare db "INSERT INTO test (a, b, c, d, e) VALUES (:a, :b, :c, :d, :e);"
      bindInt64 stmt ":a" 1
      bindDouble stmt ":b" 1.01
      bindText stmt ":c" "one"
      bindBlob stmt ":d" "\x01\x00\xf1"
      bindNull stmt ":e"
      False <- step stmt
      finalize stmt
      stmt <- prepare db "SELECT * FROM test;"
      True <- step stmt
      colCount <- columnCount stmt
      assertEqual "column count" 5 colCount
      val <- columnInt64 stmt 0
      assertEqual "integer value" 1 val
      val <- columnDouble stmt 1
      assertEqual "double value" 1.01 val
      val <- columnText stmt 2
      assertEqual "text value" "one" val
      val <- columnBlob stmt 3
      assertEqual "blob value" "\x01\x00\xf1" val
      True <- columnIsNull stmt 4
      val <- columnBlob stmt 4
      assertEqual "nullable blob value" "" val
      False <- step stmt
      finalize stmt
      close db,
    "should report an error when a statement is invalid" ~: do
      db <- open ":memory:" OpenReadWrite
      Left (ExceptionSQLite code msg) <- try $ prepare db "some garbage"
      assertEqual "error code" 1 code
      assertEqual "error message" "near \"some\": syntax error" msg
      close db,
    "should raise an error when trying to bind to non-existent parameter" ~: do
      db <- open ":memory:" OpenReadWrite
      stmt <- prepare db "SELECT 1;"
      Left (ExceptionParameterNotFound ":non-existent") <- try $ bindNull stmt ":non-existent"
      close db,
    "should raise an error when trying to use finalized statement" ~: do
      db <- open ":memory:" OpenReadWrite
      stmt <- prepare db "SELECT 1;"
      finalize stmt
      Left ExceptionStatementIsFinalized <- try $ step stmt
      close db
  ]
