module TestDsl where

import Control.Exception (ErrorCall (..))
import Data.String (String)
import Data.Time (UTCTime)
import Data.Time.Format.ISO8601 (iso8601ParseM)
import qualified Data.Vector as V
import OkSqlite.Dsl
import OkSqlite.Exception
import OkStd
import System.IO.Error (userError)
import Test.HUnit

tests =
  [ "should create, evaluate, and finalize valid SQLite statements" ~: do
      db <- open ":memory:" OpenReadWrite
      exec db createTableStmt ()
      exec db insertPredefinedValuesStmt ()
      actual <- exec db selectAllFromTest ()
      assertEqual "results count" 3 (V.length actual)
      let expected =
            V.fromList
              [ (1, 1.01, "one", "\x01\x00\xf1", Nothing, parseISO8601 "2001-01-01T01:01:01Z"),
                (2, 2.02, "に", "\x02\x00\xf2", Just "test", parseISO8601 "2002-02-02T02:02:02Z"),
                (3, 3.03, "три", "\x03\x00\xf3", Nothing, parseISO8601 "2003-03-03T03:03:03Z")
              ]
      assertEqual "results" expected actual
      close db,
    "should bind parameters" ~: do
      db <- open ":memory:" OpenReadWrite
      exec db createTableStmt ()
      exec db insertBoundValuesStmt (1, 1.01, "one", "\x01\x00\xf1", Nothing, parseISO8601 "2001-01-01T01:01:01Z")
      actual <- exec db selectAllFromTest ()
      assertEqual "results count" 1 (V.length actual)
      assertEqual "results" (1, 1.01, "one", "\x01\x00\xf1", Nothing, parseISO8601 "2001-01-01T01:01:01Z") (V.head actual)
      close db,
    "should raise an error when no rows returned but a single one was expected" ~: do
      db <- open ":memory:" OpenReadWrite
      exec db createTableStmt ()
      Left ExceptionNoRowsReturned <- try $ exec db (Statement "SELECT a FROM test;" noBind (singleRow (column :: Row Int64))) ()
      close db,
    "should raise an error when binding to non-existent parameter" ~: do
      db <- open ":memory:" OpenReadWrite
      Left (ExceptionParameterNotFound ":non-existent") <- try $ exec db (Statement "SELECT 1;" (\v -> [Bind ":non-existent" (v :: Int64)]) noResult) 123
      close db,
    "should report an error when a statement is invalid" ~: do
      db <- open ":memory:" OpenReadWrite
      Left (ExceptionSQLite code msg) <- try $ exec db (Statement "some garbage" noBind noResult) ()
      assertEqual "error code" 1 code
      assertEqual "error message" "near \"some\": syntax error" msg
      close db,
    "should raise an error when timestamp could not be parsed" ~: do
      db <- open ":memory:" OpenReadWrite
      Left (ErrorCall msg) <- try $ exec db (Statement "SELECT 'garbage';" noBind (singleRow (column :: Row UTCTime))) ()
      assertEqual "error message" "error: extract: can't parse UTCTime from 'garbage'" msg
      close db,
    "should rollback changes if transaction errors" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- exec db countTables ()
      Left exc <- try $
        inTransaction db False $ \db -> do
          exec db (Statement "CREATE TABLE foo (foo INTEGER);" noBind noResult) ()
          exec db (Statement "INSERT INTO foo (foo) VALUES (123);" noBind noResult) ()
          123 <- exec db (Statement "SELECT foo FROM foo LIMIT 1;" noBind (singleRow (column :: Row Int64))) ()
          throwM (userError "An error")
      assertEqual "transaction exception" (userError "An error") exc
      0 <- exec db countTables ()
      close db,
    "should commit changes if transaction is successful" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- exec db countTables ()
      inTransaction db False $ \db -> do
        exec db (Statement "CREATE TABLE foo (foo INTEGER);" noBind noResult) ()
        exec db (Statement "INSERT INTO foo (foo) VALUES (123);" noBind noResult) ()
      123 <- exec db (Statement "SELECT foo FROM foo LIMIT 1;" noBind (singleRow (column :: Row Int64))) ()
      close db
  ]

createTableStmt = Statement sql noBind noResult
  where
    sql = "CREATE TABLE test (a INTEGER, b REAL, c TEXT, d BLOB, e BLOB, ts TEXT);"

insertPredefinedValuesStmt = Statement sql noBind noResult
  where
    sql =
      "INSERT INTO test \
      \ (a, b, c, d, e, ts) \
      \ VALUES \
      \ (1, 1.01, 'one', x'0100f1', NULL, '2001-01-01 01:01:01'), \
      \ (2, 2.02, 'に',  x'0200f2', 'test', '2002-02-02 02:02:02'), \
      \ (3, 3.03, 'три', x'0300f3', NULL, '2003-03-03 03:03:03') \
      \ ;"

insertBoundValuesStmt :: Statement (Int64, Double, Text, ByteString, Maybe ByteString, UTCTime) ()
insertBoundValuesStmt = Statement sql binds noResult
  where
    sql = "INSERT INTO test (a, b, c, d, e, ts) VALUES (:a, :b, :c, :d, :e, :ts);"
    binds (a, b, c, d, e, ts) = [Bind ":a" a, Bind ":b" b, Bind ":c" c, Bind ":d" d, Bind ":e" e, Bind ":ts" ts]

selectAllFromTest :: Statement () (Vector (Int64, Double, Text, ByteString, Maybe ByteString, UTCTime))
selectAllFromTest = Statement sql noBind result
  where
    sql = "SELECT a, b, c, d, e, ts FROM test;"
    result = rowVector row
    row = (,,,,,) <$> (column :: Row Int64) <*> (column :: Row Double) <*> (column :: Row Text) <*> (column :: Row ByteString) <*> (column :: Row (Maybe ByteString)) <*> (column :: Row UTCTime)

countTables =
  Statement "SELECT COUNT(*) FROM sqlite_schema;" noBind (singleRow (column :: Row Int64))

parseISO8601 :: String -> UTCTime
parseISO8601 s =
  case iso8601ParseM s of
    Nothing -> error $ "can't parse ISO8601 timestamp from: " <> toText s
    Just t -> t
