module TestMigration where

import qualified Data.Vector as V
import OkSqlite.Dsl
import OkSqlite.Exception
import OkSqlite.Migration
import OkStd
import Test.HUnit

tests =
  [ "should create schema table when `migrate` is first invoked" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- countTables db
      migrate db V.empty
      1 <- countTables db
      exec db (Statement "SELECT * FROM dbschema;" noBind noResult) ()
      close db,
    "should not create schema table when `getCurrentVersion` is invoked on empty DB" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- countTables db
      0 <- getCurrentVersion db
      0 <- countTables db
      close db,
    "should apply all migraitons in order" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- getCurrentVersion db
      migrate db $ V.fromList [createTableFooStmt, insertFoo1Stmt, insertFoo2Stmt, insertFoo3Stmt]
      4 <- getCurrentVersion db
      rows <- exec db selectFooWithRowID ()
      let expected = V.fromList [(1, "one"), (2, "two"), (3, "three")]
      assertEqual "test rows" expected rows
      close db,
    "should rollback changes when migration fails" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- countTables db
      0 <- getCurrentVersion db
      Left (ExceptionSQLite _ _) <-
        try $
          migrate db $
            V.fromList
              [createTableFooStmt, createTableBarStmt, invalidStmt, createTableBazStmt]
      0 <- getCurrentVersion db
      1 <- countTables db -- Only `dbschema` table should be created
      exec db (Statement "SELECT * FROM dbschema;" noBind noResult) ()
      close db,
    "should not apply the same migration twice" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- getCurrentVersion db
      migrate db $ V.fromList [createTableFooStmt, insertFoo1Stmt]
      2 <- getCurrentVersion db
      rows <- exec db selectFooWithRowID ()
      assertEqual "test rows, 1st migration" (V.fromList [(1, "one")]) rows
      migrate db $ V.fromList [createTableFooStmt, insertFoo1Stmt, insertFoo2Stmt]
      3 <- getCurrentVersion db
      rows <- exec db selectFooWithRowID ()
      assertEqual "test rows, 2nd migration" (V.fromList [(1, "one"), (2, "two")]) rows
      close db,
    "should raise an error on mismatch between same version migrations in database and in code" ~: do
      db <- open ":memory:" OpenReadWrite
      0 <- getCurrentVersion db
      migrate db $ V.fromList [createTableFooStmt, createTableBarStmt]
      2 <- getCurrentVersion db
      Left (ExceptionMigrationCodeMismatch _ _) <- try $ migrate db $ V.fromList [createTableFooStmt, createTableBazStmt, insertFoo1Stmt]
      2 <- getCurrentVersion db
      close db
  ]

createTableFooStmt = Statement "CREATE TABLE foo (foo TEXT);" noBind noResult

createTableBarStmt = Statement "CREATE TABLE bar (bar REAL);" noBind noResult

createTableBazStmt = Statement "CREATE TABLE baz (baz INTEGER);" noBind noResult

insertFoo1Stmt = Statement "INSERT INTO foo (foo) VALUES ('one');" noBind noResult

insertFoo2Stmt = Statement "INSERT INTO foo (foo) VALUES ('two');" noBind noResult

insertFoo3Stmt = Statement "INSERT INTO foo (foo) VALUES ('three');" noBind noResult

selectFooWithRowID = Statement "SELECT rowid, foo FROM foo;" noBind (rowVector $ (,) <$> (column :: Row Int64) <*> (column :: Row Text))

invalidStmt = Statement "SELECT * FROM non_existent_table_this_should_fail;" noBind noResult

countTables db =
  exec db (Statement "SELECT COUNT(*) FROM sqlite_schema;" noBind (singleRow (column :: Row Int64))) ()
