module OkStd.Encoding.Base64Url
  ( encode,
    decode,
  )
where

import Data.Bits
import qualified Data.ByteString as BS
import qualified Data.Text as T
import Prelude

-- | Encodes binary data using Base64 URL-safe encoding without padding
encode :: BS.ByteString -> T.Text
encode buf = go $ BS.splitAt 3 buf
  where
    go (cur, rest)
      | BS.length cur == 0 =
        T.empty
      | BS.length cur == 1 =
        let i1 = shiftR (BS.index cur 0) 2
            i2 = shiftL (BS.index cur 0) 4
         in T.pack [enc i1, enc i2]
      | BS.length cur == 2 =
        let i1 = shiftR (BS.index cur 0) 2
            i2 = shiftL (BS.index cur 0) 4 .|. shiftR (BS.index cur 1) 4
            i3 = shiftL (BS.index cur 1) 2
         in T.pack [enc i1, enc i2, enc i3]
      | otherwise =
        let i1 = shiftR (BS.index cur 0) 2
            i2 = shiftL (BS.index cur 0) 4 .|. shiftR (BS.index cur 1) 4
            i3 = shiftL (BS.index cur 1) 2 .|. shiftR (BS.index cur 2) 6
            i4 = BS.index cur 2
            chunk = T.pack [enc i1, enc i2, enc i3, enc i4]
         in T.append chunk $ go $ BS.splitAt 3 rest
    enc i =
      T.index alphabet $ fromIntegral $ i .&. 0x3F

-- | Decodes binary data from Base64 URL-safe string without padding
decode :: T.Text -> Either String BS.ByteString
decode str = go $ T.splitAt 4 str
  where
    go (cur, rest)
      | T.length cur == 0 =
        Right BS.empty
      | T.length cur == 1 =
        Left "Unexpected EOF"
      | T.length cur == 2 = do
        i1 <- dec $ T.index cur 0
        i2 <- dec $ T.index cur 1
        pure $ BS.singleton $ shiftL i1 2 .|. shiftR i2 4
      | T.length cur == 3 = do
        i1 <- dec $ T.index cur 0
        i2 <- dec $ T.index cur 1
        i3 <- dec $ T.index cur 2
        pure $
          BS.pack
            [ shiftL i1 2 .|. shiftR i2 4,
              shiftL i2 4 .|. shiftR i3 2
            ]
      | otherwise = do
        i1 <- dec $ T.index cur 0
        i2 <- dec $ T.index cur 1
        i3 <- dec $ T.index cur 2
        i4 <- dec $ T.index cur 3
        let chunk =
              BS.pack
                [ shiftL i1 2 .|. shiftR i2 4,
                  shiftL i2 4 .|. shiftR i3 2,
                  shiftL i3 6 .|. i4
                ]
        BS.append chunk <$> go (T.splitAt 4 rest)
    dec x =
      case T.findIndex (== x) alphabet of
        Just i -> Right $ fromIntegral i
        Nothing -> Left $ "Invalid character: '" <> [x] <> "'"

alphabet :: T.Text
alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"