{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE TemplateHaskell #-}

module Main where

import qualified Data.ByteString as BS
import Data.FileEmbed (embedFile)
import qualified Data.List as L
import qualified Data.Text as T
import qualified Data.Vector as V
import Data.Version (showVersion)
import qualified Network.Wai.Handler.Warp as Warp
import OkSqlite
import OkStd
import OkStd.Logger
import OkVault.Config (configKeys)
import OkVault.Context (logInfo)
import qualified OkVault.Context as Context
import OkVault.DatabaseSchema (ensureDatabaseSchema)
import OkVault.Models.Exception (ExceptionModel (..))
import qualified OkVault.Models.User as User
import qualified OkVault.Router as Router
import Paths_okvault (version)
import System.Exit (exitWith)
import System.IO (hClose)
import System.Process (CreateProcess (..), StdStream (CreatePipe), proc, waitForProcess, withCreateProcess)

main = do
  dbPath <- fromMaybe "okvault.db" <$> lookupEnv "DBPATH"
  getArgs >>= \case
    ["serve"] -> cmdServe
    ["init"] -> cmdInit dbPath
    ["show-config"] -> cmdShowConfig dbPath
    ["set-config-option", key, value] -> cmdSetConfigOption dbPath key value
    ["invite", email] -> cmdInvite dbPath email
    ["list-users"] -> cmdListUsers dbPath
    ["help"] -> cmdHelp
    ["version"] -> cmdVersion
    [] -> cmdHelp
    args -> cmdNotFound $ unwords args

cmdHelp = do
  exitCode <- withCreateProcess manProc $ \(Just stdin) _ _ p -> do
    BS.hPut stdin manPage
    hClose stdin
    waitForProcess p
  exitWith exitCode
  where
    manProc =
      (proc "groff" ["-Tascii", "-man", "-"]) {delegate_ctlc = True, std_in = CreatePipe}
    manPage =
      $(embedFile "man/okvault.1")

cmdNotFound cmd =
  error $ "unknown command: '" <> cmd <> "'" <> "\n(use 'okvault help' for a list of commands)"

cmdVersion =
  putText $ "OkVault " <> toText (showVersion version)

cmdServe = do
  ctx <- Context.fromDatabase
  let ?ctx = ctx
  let settings =
        [ Warp.setPort (Context.getPort ctx),
          Warp.setHost (fromString $ Context.getHost ctx),
          Warp.setBeforeMainLoop $
            logInfo $
              "OkVault " <> toLogStr (showVersion version)
                <> " is listening on http://"
                <> toLogStr (Context.getHost ctx)
                <> ":"
                <> toLogStr (Context.getPort ctx)
        ]
  Warp.runSettings (foldl' (flip ($)) Warp.defaultSettings settings) Router.routeRequest

cmdInit dbPath =
  withDatabase dbPath OpenReadWriteCreate ensureDatabaseSchema

cmdShowConfig dbPath =
  withDatabase dbPath OpenReadOnly $ \db -> do
    for_ configKeys $ \(k, (def, desc)) -> do
      cur <- readKeyDefault db k def
      putStrLn $ "# " <> desc
      putStrLn $ k <> " = '" <> cur <> "'\n"
  where
    readKeyDefault db key def =
      fromMaybe def
        <$> exec
          db
          ( Statement
              "SELECT value FROM config WHERE key = :key;"
              (const [Bind ":key" key])
              (rowMaybe (column :: Row Text))
          )
          ()

cmdSetConfigOption dbPath key value =
  case lookup key configKeys of
    Nothing ->
      error $ "invalid configuration option: '" <> key <> "'"
    Just _ -> do
      withDatabase dbPath OpenReadWrite $ \db ->
        exec
          db
          ( Statement
              "INSERT OR REPLACE INTO config (key, value) VALUES (:key, :value);"
              (const [Bind ":key" key, Bind ":value" value])
              noResult
          )
          ()
      putText "Done"

cmdInvite dbPath email =
  withDatabase dbPath OpenReadWrite (User.create email)
    `catches` [ Handler $ \case
                  ExceptionSQLite _ msg -> error msg
                  exc -> error $ toText $ show exc,
                Handler $ \case
                  ExceptionAlreadyExists -> error "user is already invited"
              ]

cmdListUsers dbPath = do
  users <- withDatabase dbPath OpenReadOnly User.findAll
  printTable ["ID", "EMAIL", "INVITED AT"] $
    V.toList $
      users <&> \User.User {..} ->
        [toText (show id), email, toText (show createdAt)]

printTable :: [Text] -> [[Text]] -> IO ()
printTable header rows =
  for_ allRows $ \row -> do
    for_ (L.zipWith padWithSpaces columnWidths row) putStr
    putText ""
  where
    allRows = header : rows
    columnWidths =
      let go [] = L.replicate (L.length header) 0
          go (r : rs) = L.zipWith max (fmap T.length r) $ go rs
       in go allRows
    padWithSpaces width value =
      value <> T.replicate (width - T.length value + 3) " "
