module OkVault.Actions.Auth
  ( showLoginPage,
    sendLoginCode,
    validateLoginCode,
    showLogoutPage,
    logOut,
  )
where

import Data.Time (NominalDiffTime, addUTCTime, getCurrentTime)
import Data.Time.Clock.POSIX (posixSecondsToUTCTime)
import qualified Network.HTTP.Types as HTTP
import OkStd
import OkStd.Logger
import OkVault.Actions.Error
import OkVault.Actions.Shared
import OkVault.Context
import qualified OkVault.Models.LoginCode as LoginCode
import qualified OkVault.Models.Session as Session
import qualified OkVault.Models.User as User
import qualified OkVault.Templates as Templates
import Web.Request (createCookie, parseForm, remoteHost)

loginCodeExpirationSeconds :: NominalDiffTime
loginCodeExpirationSeconds = 30 * 60 -- 30 minutes

sessionExpirationSeconds :: NominalDiffTime
sessionExpirationSeconds = 30 * 24 * 60 * 60 -- 30 days

showLoginPage :: Action
showLoginPage =
  currentUser >>= \case
    Just _ -> redirectTemporarily "/"
    Nothing -> showHTML HTTP.status200 Templates.loginPage

sendLoginCode :: Action
sendLoginCode = do
  maybeEmail <- lookup "email" <$> parseForm ?req
  case maybeEmail of
    Nothing ->
      showMalformedRequestPage
    Just email -> do
      logInfo $ "Login attempt: email='" <> toLogStr email <> "' from='" <> toLogStr (remoteHost ?req) <> "'"
      void $
        withDB $ \db -> runMaybeT $ do
          u <- MaybeT $ User.findByEmail email db
          codeExpiresAt <- lift $ addUTCTime loginCodeExpirationSeconds <$> getCurrentTime
          code <- lift $ LoginCode.create (User.id u) codeExpiresAt db
          lift $ sendCode email code
      pure (HTTP.status200, [], Templates.loginCodePage email)
  where
    sendCode email code =
      let subject = code <> " -- your OkVault login code"
          body =
            unlines
              [ "Dear user,",
                "",
                "Your OkVault login code is " <> code,
                "",
                "Copy/paste or enter it into the login page you should be seeing right now.",
                "",
                "Cheers,",
                "OkVault"
              ]
       in sendMail email subject body

validateLoginCode :: Action
validateLoginCode = do
  params <- parseForm ?req <&> \p -> sequence [lookup "email" p, lookup "code" p]
  case params of
    Just [email, codeValue] -> do
      maybeSession <- withDB $ \db -> runMaybeT $ do
        u <- MaybeT $ User.findByEmail email db
        c <- MaybeT $ LoginCode.findValidByValueAndUserID codeValue (User.id u) db
        lift $ LoginCode.invalidate c db
        now <- lift getCurrentTime
        let sessionExpiresAt = addUTCTime sessionExpirationSeconds now
        lift $ Session.create (User.id u) sessionExpiresAt db
      case maybeSession of
        Nothing -> do
          logInfo $ "Access denied: email='" <> toLogStr email <> "' from='" <> toLogStr (remoteHost ?req) <> "'"
          showErrorPage HTTP.status403 "Authentication Failed" "We could not authenticate you."
        Just session -> do
          logInfo $ "Access granted: email='" <> toLogStr email <> "' from='" <> toLogStr (remoteHost ?req) <> "'"
          let sidCookie = createCookie "okvault.sid" (encodeUtf8 $ Session.id session) (Session.expiresAt session) (not $ getAllowInsecure ?ctx)
          (s, hs, c) <- redirectTemporarily "/"
          pure (s, sidCookie : hs, c)
    _ ->
      showMalformedRequestPage

showLogoutPage :: Action
showLogoutPage =
  whenAuthenticated $ \_ _ ->
    showHTML HTTP.status200 Templates.logoutPage

logOut :: Action
logOut =
  whenAuthenticated $ \session _ -> do
    withDB (Session.delete session)
    (s, hs, c) <- redirectTemporarily "/"
    pure (s, clearSidCookie : hs, c)
  where
    clearSidCookie = createCookie "okvault.sid" "" (posixSecondsToUTCTime 0) (not $ getAllowInsecure ?ctx)
