module OkVault.Actions.Item
  ( showNewItemPage,
    createNewItem,
    showItem,
    showEditItemPage,
    updateItem,
    showDeleteItemPage,
    deleteItem,
  )
where

import qualified Network.HTTP.Types as HTTP
import OkStd
import OkVault.Actions.Error
import OkVault.Actions.Shared
import OkVault.Context
import qualified OkVault.Models.Item as Item
import qualified OkVault.Models.Session as Session
import qualified OkVault.Models.User as User
import qualified OkVault.Templates as Templates

showNewItemPage :: Action
showNewItemPage =
  whenAuthenticatedAndSet $ \s u ->
    showHTML HTTP.status200 $ Templates.addEditItem u Nothing (Session.csrfToken s)

createNewItem :: Action
createNewItem =
  whenAuthenticatedAndSet $ \_ u ->
    whenCSRFTokenValid $ \params ->
      case sequence
        [ lookup "name" params,
          lookup "category" params,
          lookup "content" params,
          lookup "item_key" params
        ] of
        Just [name, category, content, itemKey] -> do
          item <- withDB (Item.create (User.id u) name category content itemKey)
          redirectTemporarily $ "/item/" <> Item.slug item
        _ -> do
          logDebug "createNewItem: missing parameters"
          showMalformedRequestPage

showItem :: Text -> Action
showItem slug =
  whenAuthenticatedAndSet $ \_ u ->
    withDB (Item.findBySlugAndUserID slug (User.id u)) >>= \case
      Nothing ->
        showNotFoundPage
      Just item ->
        showHTML HTTP.status200 $ Templates.showItem u item

showEditItemPage :: Text -> Action
showEditItemPage slug =
  whenAuthenticatedAndSet $ \s u ->
    withDB (Item.findBySlugAndUserID slug (User.id u)) >>= \case
      Nothing ->
        showNotFoundPage
      Just item ->
        showHTML HTTP.status200 $ Templates.addEditItem u (Just item) (Session.csrfToken s)

updateItem :: Text -> Action
updateItem slug =
  whenAuthenticatedAndSet $ \_ u ->
    whenCSRFTokenValid $ \params ->
      case sequence
        [ lookup "name" params,
          lookup "category" params,
          lookup "content" params,
          lookup "item_key" params
        ] of
        Just [name, category, content, itemKey] -> do
          withDB (Item.findBySlugAndUserID slug (User.id u)) >>= \case
            Nothing -> do
              logDebug "updateItem: item does not exist"
              showMalformedRequestPage
            Just _ -> do
              withDB (Item.update slug (User.id u) name category content itemKey)
              redirectTemporarily $ "/item/" <> slug
        _ -> do
          logDebug "updateItem: missing parameters"
          showMalformedRequestPage

showDeleteItemPage :: Text -> Action
showDeleteItemPage slug =
  whenAuthenticatedAndSet $ \s u ->
    withDB (Item.findBySlugAndUserID slug (User.id u)) >>= \case
      Nothing ->
        showNotFoundPage
      Just item ->
        showHTML HTTP.status200 $ Templates.deleteItem u item (Session.csrfToken s)

deleteItem :: Text -> Action
deleteItem slug =
  whenAuthenticatedAndSet $ \_ u ->
    whenCSRFTokenValid $ \_ ->
      withDB (Item.findBySlugAndUserID slug (User.id u)) >>= \case
        Nothing -> do
          logDebug "deleteItem: item does not exist"
          showMalformedRequestPage
        Just _ -> do
          withDB (Item.archive slug (User.id u))
          redirectTemporarily "/"
