module OkVault.Actions.Shared
  ( Action,
    showHTML,
    showErrorPage,
    redirectTemporarily,
    currentSession,
    currentUser,
    whenAuthenticated,
    whenAuthenticatedAndSet,
    whenCSRFTokenValid,
  )
where

import Data.Time (getCurrentTime)
import Network.HTTP.Types (ResponseHeaders, Status)
import qualified Network.HTTP.Types as HTTP
import Network.Wai (Request)
import OkStd
import OkVault.Context
import OkVault.Models.Session (Session)
import qualified OkVault.Models.Session as Session
import OkVault.Models.User (User)
import qualified OkVault.Models.User as User
import qualified OkVault.Templates as Templates
import Web.Request (parseCookies, parseForm)

-- | Type synonym for OkVault HTTP handler operating in application context
-- with specified HTTP request
type Action = (?ctx :: Context, ?req :: Request) => IO (Status, ResponseHeaders, ByteString)

-- | Type synonym for a function operating inside OkVault HTTP handler and
-- using application context and/or HTTP request
type InAction a = (?ctx :: Context, ?req :: Request) => IO a

-- | Returns specified HTTP status and HTML content
showHTML :: Status -> ByteString -> Action
showHTML status content =
  pure (status, [("Content-Type", "text/html; charset=UTF-8")], content)

-- | Returns specified HTTP status and an error page rendered with the given
-- title and message
showErrorPage :: Status -> Text -> Text -> Action
showErrorPage status title msg =
  showHTML status $ Templates.errorPage title msg

redirectTemporarily :: Text -> Action
redirectTemporarily url =
  pure (HTTP.status303, [("Location", encodeUtf8 url)], "")

-- | Returns session of the currently logged in user. Does not check if the
-- user still exists.
currentSession :: InAction (Maybe Session)
currentSession = runMaybeT $ do
  sid <- hoistMaybe $ lookup "okvault.sid" $ parseCookies ?req
  s <- MaybeT $ withDB (Session.findByID sid)
  now <- lift getCurrentTime
  guard (Session.expiresAt s > now)
  pure s

-- | Returns user that is currently logged in
currentUser :: InAction (Maybe User)
currentUser =
  currentSession >>= \case
    Just s ->
      withDB (User.findByID $ Session.userID s)
    Nothing ->
      pure Nothing

-- | Checks current session and executes specified action if user exists and is
-- authenticated. Redirects to the login page otherwise.
whenAuthenticated :: (Session -> User -> Action) -> Action
whenAuthenticated action =
  currentSessionAndUser >>= \case
    Just (s, u) -> action s u
    Nothing -> redirectTemporarily "/auth/login"
  where
    currentSessionAndUser = runMaybeT $ do
      s <- MaybeT currentSession
      u <- MaybeT $ withDB (User.findByID $ Session.userID s)
      pure (s, u)

-- | Checks current session and executes specified action if user exists, is
-- authenticated, and has all necessary keys set. Redirects to login page if
-- user does not exist or not authenticated. Redirects to the first time page if
-- one or more user keys are missing.
whenAuthenticatedAndSet :: (Session -> User -> Action) -> Action
whenAuthenticatedAndSet action =
  whenAuthenticated $ \s u ->
    if User.hasKeys u
      then action s u
      else redirectTemporarily "/first-time"

-- | Checks current session and executes specified action if user is
-- authenticated and `csrf_token` parameter of the request body matches session
-- CSRF token. Returns 403 if token is invalid. Redirects to the login page if
-- not authenticated.
whenCSRFTokenValid :: ([(Text, Text)] -> Action) -> Action
whenCSRFTokenValid action =
  whenAuthenticated $ \s _ -> do
    params <- parseForm ?req
    if fromMaybe "" (lookup "csrf_token" params) /= Session.csrfToken s
      then showErrorPage HTTP.status403 "Invalid security token" "Go back, refresh the page, and try again please"
      else action params
