module OkVault.Context
  ( Context,
    InContext,
    fromDatabase,
    getDBPath,
    getHost,
    getPort,
    getBaseURL,
    getAllowInsecure,
    getSendmail,
    getEmailAddress,
    getTemplates,
    getBundledCSS,
    getBundledJS,
    withDB,
    sendMail,
    logInfo,
    logDebug,
  )
where

import qualified Data.List.NonEmpty as NonEmpty
import Data.String (String)
import Data.Text (unpack)
import Data.Text.Lazy (fromStrict)
import OkSqlite
import OkStd
import OkStd.Logger
import OkVault.Config (configKeys)
import qualified OkVault.Resources as Resources
import System.Process (readProcess)
import Text.Microstache
  ( MustacheException (..),
    Template,
    compileMustacheText,
  )
import Text.Read (readMaybe)

-- | Application context. Contains all options available via environment
-- variables.
data Context = Context
  { getDBPath :: Text,
    getHost :: String,
    getPort :: Int,
    getBaseURL :: Text,
    getAllowInsecure :: Bool,
    getSendmail :: Text,
    getEmailAddress :: Text,
    getDebugFlag :: Bool,
    getTemplates :: Template,
    getBundledCSS :: (Text, ByteString),
    getBundledJS :: (Text, ByteString)
  }
  deriving (Eq, Show)

-- | Type synonym for a function that needs access to application context
type InContext a = (?ctx :: Context) => a

-- | Creates new application context by reading configuration from the
-- database. Path to the database is read from DBPATH environment variable or
-- 'okvault.db' if the variable is unset.
fromDatabase :: IO Context
fromDatabase = do
  dbPath <- fromMaybe "okvault.db" <$> lookupEnv "DBPATH"
  bundledCSS@(bundledCssFilePath, _) <- Resources.bundleCSS
  bundledJS@(bundledJsFilePath, _) <- Resources.bundleJS
  withDatabase dbPath OpenReadWrite $ \db ->
    Context dbPath
      <$> (unpack <$> readKey db "host")
      <*> (parsePort <$> readKey db "port")
      <*> readKey db "base-url"
      <*> ((== "yes") <$> readKey db "allow-insecure")
      <*> readKey db "sendmail"
      <*> readKey db "email-address"
      <*> ((== "yes") <$> readKey db "debug")
      <*> compileTemplates bundledCssFilePath bundledJsFilePath
      <*> pure bundledCSS
      <*> pure bundledJS
  where
    parsePort v =
      let v' = unpack v
       in case readMaybe v' of
            Nothing -> error $ "invalid port: '" <> v <> "'"
            Just p -> p
    readKey :: Database -> Text -> IO Text
    readKey db key =
      case lookup key configKeys of
        Nothing ->
          error $ "configuration key is not defined in OkVault.Config: " <> key
        Just (def, _) ->
          fromMaybe def
            <$> exec
              db
              ( Statement
                  "SELECT value FROM config WHERE key = :key;"
                  (const [Bind ":key" key])
                  (rowMaybe (column :: Row Text))
              )
              ()
    compileTemplates bundledCssFileName bundledJsFileName = do
      namesAndFiles <- Resources.listTemplates
      templates <- for (NonEmpty.fromList namesAndFiles) $ \(name, path) -> do
        code <- decodeUtf8 <$> Resources.getResource path
        compile' name code
      cssTemplate <-
        compile'
          "bundled_css"
          ("<link rel=\"stylesheet\" type=\"text/css\" href=\"/static/" <> bundledCssFileName <> "\"/>")
      jsTemplate <-
        compile'
          "bundled_js"
          ("<script src=\"/static/" <> bundledJsFileName <> "\" defer></script>")
      pure $ sconcat templates <> cssTemplate <> jsTemplate
      where
        compile' name code =
          case compileMustacheText name $ fromStrict code of
            Left parseError -> throwM $ MustacheParserException parseError
            Right template -> pure template

-- | Creates new connection to the database under DBPATH and runs specified
-- action with it
withDB :: (Database -> IO a) -> InContext (IO a)
{-# INLINE withDB #-}
withDB = withDatabase (getDBPath ?ctx) OpenReadWrite

-- | Sends mail by invoking SENDMAIL program. Raises `IOError` when SENDMAIL
-- fails.
sendMail :: Text -> Text -> Text -> InContext (IO ())
sendMail to subject body =
  void $ readProcess sendmailExe [toArg] msg
  where
    sendmailExe = unpack $ getSendmail ?ctx
    toArg = unpack to
    from = getEmailAddress ?ctx
    msg =
      unpack $
        unlines
          [ "To: " <> to,
            "From: OkVault <" <> from <> ">",
            "Subject: " <> subject,
            "",
            body
          ]

logInfo :: LogStr -> InContext (IO ())
{-# INLINE logInfo #-}
logInfo msg = logM ("[INF] " <> msg)

logDebug :: LogStr -> InContext (IO ())
{-# INLINE logDebug #-}
logDebug msg =
  if getDebugFlag ?ctx then logM ("[DBG] " <> msg) else pure ()
