Database Schema
===============

This module contains database schema of OkVault.

Encrypted Data Format (EDF)
---------------------------

All encrypted data is stored in the following format:

- 8 bytes of `encdat01` magic string
- 24 bytes of randomly chosen nonce value used for encryption
- ciphertext

The whole payload is encoded using Base64 URL-safe encoding.

Note on database vs application types
-------------------------------------

Even though BLOB is used in database schema to indicate encrypted binary data
in EDF, Text is used throughout Haskell code because it's more convenient. It
also indicates that the data must be a valid UTF-8 string because it's encoded
in Base64.

\begin{code}
module OkVault.DatabaseSchema
    ( ensureDatabaseSchema
    ) where

import OkStd
import OkSqlite
import qualified Data.Vector as V

ensureDatabaseSchema :: Database -> IO ()
ensureDatabaseSchema db =
    migrate db $ V.fromList migrations
  where
    migrations = (\sql -> Statement sql noBind noResult) <$>
\end{code}

`config` table stores OkVault configuration.

\begin{code}
        [ "CREATE TABLE config \
            \( key TEXT NOT NULL PRIMARY KEY \
            \, value TEXT NOT NULL \
            \);"
\end{code}

`users` table stores emails and encryption keys of OkVault users.

Administrators can invite new users using `okvaultadm` tool. It creates new
record in the `users` table leaving `data_key` and `metadata_key` fields empty.
Values for these fields are generated on the user's computer upon first login
following this algorithm:

1. Ask user to enter their Master Password (MsP)
2. Generate random salt (S) for derivation
3. Derive Master Key (MsK) from MP using S
4. Generate new random Data Key (DK) and Metadata Key (MK)
5. Encrypt DK and MK using MsK
6. Send encrypted DK, MK, and S to the server

\begin{code}
        , "CREATE TABLE users \
            \( id INTEGER NOT NULL PRIMARY KEY \
            \, email TEXT NOT NULL UNIQUE \
            \, salt BLOB \
            \, data_key BLOB \
            \, metadata_key BLOB \
            \, created_at TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP \
            \);"
\end{code}

`login_codes` tables stores temporary codes that are sent via email to the user
to confirm their login.

`code` is a randomly generated short alphanumeric string that is valid as a
login proof until `expires_at`. `expires_at` should be set to a time in the
near future, 5 or 10 minutes max.

\begin{code}
        , "CREATE TABLE login_codes \
            \( value TEXT NOT NULL \
            \, user_id INTEGER NOT NULL \
            \, expires_at TEXT NOT NULL \
            \, created_at TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP \
            \);"
\end{code}

`sessions` contains IDs and expiration times of user session. A session is created
upon successful login and its ID is stored in site cookies. Session expiration
time could be quite long, up to 30 days.

\begin{code}
        , "CREATE TABLE sessions \
            \( id TEXT NOT NULL PRIMARY KEY \
            \, csrf_token TEXT NOT NULL \
            \, user_id INTEGER NOT NULL \
            \, created_at TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP \
            \, expires_at TEXT NOT NULL \
            \);"
\end{code}

`items` is where logins and passwords, bank cards, and other secrets are
stored.

Each item has a `slug` which is a short alphanumeric label. Together with
`user_id` it is used to uniquely refer to the item in a user-friendly way.

`name` and `category` are the name and category of the item encrypted with
MK. `content` is the content of the item encrypted with the Item Key (IK). IK
is stored in `item_key` field and encrypted with DK.

`archived_at` is a timestamp of when the item was archived. For current items
it's empty string. Archived items are preserved in the database, but are not
showed in the application.

\begin{code}
        , "CREATE TABLE items \
            \( id INTEGER NOT NULL PRIMARY KEY \
            \, slug TEXT NOT NULL \
            \, user_id INTEGER NOT NULL \
            \, name BLOB NOT NULL \
            \, category BLOB NOT NULL \
            \, content BLOB NOT NULL \
            \, item_key BLOB NOT NULL \
            \, created_at TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP \
            \, archived_at TEXT NOT NULL DEFAULT '' \
            \, CONSTRAINT 'unique_user_id_slug_archived_at' UNIQUE (user_id, slug, archived_at) \
            \, CONSTRAINT 'no_empty_slug' CHECK (slug <> '') \
            \);"
        ]
\end{code}
