{-# OPTIONS_GHC -Wno-name-shadowing #-}

module OkVault.Models.Item
  ( Item (..),
    Archived (..),
    create,
    update,
    archive,
    findAllByUserID,
    findBySlugAndUserID,
  )
where

import Data.Time (UTCTime)
import OkCrypt (randomBytes)
import OkSqlite
import qualified OkSqlite.Bindings as B
import OkSqlite.Values (ExtractableValue (..))
import OkStd

data Archived
  = NotArchived
  | ArchivedAt UTCTime
  deriving (Eq, Show)

instance ExtractableValue Archived where
  extract s c =
    B.columnText s c >>= \case
      "" -> pure NotArchived
      _ -> ArchivedAt <$!> extract s c

-- | Item is a model containing encrypted data and metadata of an OkVault
-- item, e.g. login credentials, bank card details, secret note, etc.
data Item = Item
  { id :: Int64,
    slug :: Text,
    userID :: Int64,
    name :: Text,
    category :: Text,
    content :: Text,
    itemKey :: Text,
    createdAt :: UTCTime,
    archivedAt :: Archived
  }
  deriving (Eq, Show)

-- | Creates new item for the specified user
create :: Int64 -> Text -> Text -> Text -> Text -> Database -> IO Item
create userID name category content itemKey db = do
  slug <- toHexString <$> randomBytes 16
  exec db (Statement sql binds noResult) slug
  findBySlugAndUserID slug userID db >>= maybe (bug "Item.create: can't find newly created item") pure
  where
    sql =
      "INSERT INTO items \
      \ (slug, user_id, name, category, content, item_key) \
      \ VALUES \
      \ (:slug, :user_id, :name, :category, :content, :item_key) \
      \;"
    binds slug =
      [ Bind ":slug" slug,
        Bind ":user_id" userID,
        Bind ":name" name,
        Bind ":category" category,
        Bind ":content" content,
        Bind ":item_key" itemKey
      ]

-- | Creates new version of the item under specified slug. Does nothing if
-- item does not exist.
update :: Text -> Int64 -> Text -> Text -> Text -> Text -> Database -> IO ()
update slug userID name category content itemKey db = do
  archive slug userID db
  exec db (Statement sql binds noResult) ()
  where
    sql =
      "INSERT INTO items \
      \   (slug, user_id, name, category, content, item_key) \
      \ SELECT \
      \   :slug, :user_id, :name, :category, :content, :item_key \
      \ WHERE EXISTS ( \
      \   SELECT id FROM items WHERE slug = :slug AND user_id = :user_id \
      \ );"
    binds () =
      [ Bind ":slug" slug,
        Bind ":user_id" userID,
        Bind ":name" name,
        Bind ":category" category,
        Bind ":content" content,
        Bind ":item_key" itemKey
      ]

-- | Archives the item under specified slug. Does nothing if item does not
-- exist.
archive :: Text -> Int64 -> Database -> IO ()
archive slug userID db = do
  exec db (Statement sql binds noResult) ()
  where
    sql =
      "UPDATE items \
      \ SET archived_at = CURRENT_TIMESTAMP \
      \ WHERE slug = :slug \
      \   AND user_id = :user_id \
      \   AND archived_at = '' \
      \;"
    binds () =
      [ Bind ":slug" slug,
        Bind ":user_id" userID
      ]

-- | Finds all current (i.e. not archived yet) items for the specified user
findAllByUserID :: Int64 -> Database -> IO (Vector Item)
findAllByUserID userID db =
  exec db (Statement sql binds $ rowVector itemRow) ()
  where
    sql =
      "SELECT \
      \ id, slug, user_id, name, category, content, item_key, created_at, archived_at \
      \ FROM items \
      \ WHERE user_id = :user_id \
      \   AND archived_at = '' \
      \;"
    binds () = [Bind ":user_id" userID]

-- | Find non-archived item by its slug and user ID. Returns nothing if item
-- does not exist or is archived.
findBySlugAndUserID :: Text -> Int64 -> Database -> IO (Maybe Item)
findBySlugAndUserID slug userID db =
  exec db (Statement sql binds $ rowMaybe itemRow) ()
  where
    sql =
      "SELECT \
      \ id, slug, user_id, name, category, content, item_key, created_at, archived_at \
      \ FROM items \
      \ WHERE slug = :slug \
      \   AND user_id = :user_id \
      \   AND archived_at = '' \
      \;"
    binds () = [Bind ":slug" slug, Bind ":user_id" userID]

itemRow :: Row Item
itemRow =
  Item
    <$> (column :: Row Int64)
    <*> (column :: Row Text)
    <*> (column :: Row Int64)
    <*> (column :: Row Text)
    <*> (column :: Row Text)
    <*> (column :: Row Text)
    <*> (column :: Row Text)
    <*> (column :: Row UTCTime)
    <*> (column :: Row Archived)
