{-# OPTIONS_GHC -Wno-name-shadowing #-}

module OkVault.Models.LoginCode
  ( LoginCode (..),
    create,
    findValidByValueAndUserID,
    invalidate,
  )
where

import qualified Data.ByteString as BS
import qualified Data.Text as T
import Data.Time (UTCTime)
import OkCrypt (randomBytes)
import OkSqlite
import OkStd

-- | Login code is a model for a short randomly generated string used to
-- authenticate users via email.
data LoginCode = LoginCode
  { value :: Text,
    userID :: Int64,
    expiresAt :: UTCTime,
    createdAt :: UTCTime
  }
  deriving (Eq, Show)

-- | Creates new code for a user under given ID. Sets its expiration time to
-- the specified time. Returns values of the new login code.
create :: Int64 -> UTCTime -> Database -> IO Text
create userID expiresAt db = do
  value <- generateCode
  exec db (Statement sql binds noResult) value
  pure value
  where
    sql = "INSERT INTO login_codes (value, user_id, expires_at) VALUES (:value, :user_id, :expires_at);"
    binds value = [Bind ":value" value, Bind ":user_id" userID, Bind ":expires_at" expiresAt]
    generateCode = do
      let alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
      fmap toText $
        for ([1 .. 8] :: [Int]) $ \_ -> do
          n <- fromIntegral . BS.head <$> randomBytes 1
          pure $ T.index alphabet $ n `mod` T.length alphabet

-- | Finds the least recently created valid login code by its value and given
-- user ID. Returns nothing if login code does not exist or is expired.
findValidByValueAndUserID :: Text -> Int64 -> Database -> IO (Maybe LoginCode)
findValidByValueAndUserID value userID db =
  exec db (Statement sql binds result) ()
  where
    sql =
      "SELECT \
      \ value, user_id, expires_at, created_at \
      \ FROM login_codes \
      \ WHERE \
      \   value = :value \
      \   AND user_id = :user_id \
      \   AND expires_at > CURRENT_TIMESTAMP \
      \;"
    binds () = [Bind ":value" value, Bind ":user_id" userID]
    result =
      rowMaybe $
        LoginCode
          <$> (column :: Row Text)
          <*> (column :: Row Int64)
          <*> (column :: Row UTCTime)
          <*> (column :: Row UTCTime)

-- | Invalidates specified login code by setting its expiresAt field to
-- 1970-01-01 00:00:00
invalidate :: LoginCode -> Database -> IO ()
invalidate code db =
  exec db (Statement sql binds noResult) ()
  where
    sql =
      "UPDATE login_codes \
      \ SET expires_at = datetime(0, 'unixepoch') \
      \ WHERE \
      \   value = :value \
      \   AND user_id = :user_id \
      \;"
    binds () =
      [ Bind ":value" $ value code,
        Bind ":user_id" $ userID code
      ]
