{-# OPTIONS_GHC -Wno-name-shadowing #-}

module OkVault.Models.Session
  ( Session (..),
    create,
    delete,
    findByID,
  )
where

import Data.Time (UTCTime)
import OkCrypt (randomBytes)
import OkSqlite
import OkStd

-- | Session is a model containing information about current and past user
-- sessions, as well as a token used to protect users from CSRF attacks.
data Session = Session
  { id :: Text,
    csrfToken :: Text,
    userID :: Int64,
    createdAt :: UTCTime,
    expiresAt :: UTCTime
  }
  deriving (Eq, Show)

-- | Creates new session for a user under given ID. Sets its expiration time to
-- the given time. Returns newly created session.
create :: Int64 -> UTCTime -> Database -> IO Session
create userID expiresAt db = do
  sid <- toHexString <$> randomBytes 16
  csrfToken <- toHexString <$> randomBytes 16
  exec db (Statement sql binds noResult) (sid, csrfToken)
  findByID sid db >>= maybe (error "Session.create: can't find newly created session") pure
  where
    sql =
      "INSERT INTO sessions \
      \ (id, csrf_token, user_id, expires_at) \
      \ VALUES \
      \ (:id, :csrf_token, :user_id, :expires_at)\
      \;"
    binds (sid, csrfToken) =
      [ Bind ":id" sid,
        Bind ":csrf_token" csrfToken,
        Bind ":user_id" userID,
        Bind ":expires_at" expiresAt
      ]

-- | Deletes the session
delete :: Session -> Database -> IO ()
delete s db = do
  exec db (Statement sql binds noResult) ()
  where
    sql =
      "DELETE FROM sessions \
      \ WHERE id = :id \
      \ AND user_id = :user_id \
      \;"
    binds () = [Bind ":id" (id s), Bind ":user_id" (userID s)]

-- | Finds session by its ID
findByID :: Text -> Database -> IO (Maybe Session)
findByID id db =
  exec db (Statement sql binds result) ()
  where
    sql =
      "SELECT \
      \ id, csrf_token, user_id, created_at, expires_at \
      \ FROM sessions \
      \ WHERE id = :id \
      \;"
    binds () = [Bind ":id" id]
    result =
      rowMaybe $
        Session
          <$> (column :: Row Text)
          <*> (column :: Row Text)
          <*> (column :: Row Int64)
          <*> (column :: Row UTCTime)
          <*> (column :: Row UTCTime)
