{-# OPTIONS_GHC -Wno-name-shadowing #-}

module OkVault.Models.User
  ( User (..),
    hasKeys,
    create,
    setKeys,
    findByEmail,
    findByID,
    findAll,
  )
where

import Data.List (all)
import Data.Maybe (isJust)
import Data.Time (UTCTime)
import OkSqlite
import OkSqlite.Codes (c_SQLITE_CONSTRAINT)
import OkStd
import OkVault.Models.Exception

-- | User is a model containing information used for authenticating person
-- using the app and encrypting their data and metadata.
data User = User
  { id :: Int64,
    email :: Text,
    salt :: Maybe Text,
    dataKey :: Maybe Text,
    metadataKey :: Maybe Text,
    createdAt :: UTCTime
  }
  deriving (Eq, Show)

-- | Returns whether a user has all necessary keys set up
hasKeys :: User -> Bool
hasKeys u = all isJust [dataKey u, metadataKey u, salt u]

-- | Creates new user with the specified email in the database. Raises
-- @ExceptionAlreadyExists@ if user with the same email already exists in the
-- database.
create :: Text -> Database -> IO ()
create email db =
  exec db (Statement sql binds noResult) () `catch` \case
    ExceptionSQLite code _ | code == c_SQLITE_CONSTRAINT -> throwM ExceptionAlreadyExists
    exc -> throwM exc
  where
    sql = "INSERT INTO users (email) VALUES (:email);"
    binds () = [Bind ":email" email]

-- | Update user record with master key salt, data and metadata keys. Raises NO
-- exceptions if user does not exist.
setKeys :: Int64 -> Text -> Text -> Text -> Database -> IO ()
setKeys id salt dataKey metadataKey db =
  exec db (Statement sql binds noResult) ()
  where
    sql =
      "UPDATE users \
      \ SET salt = :salt, \
      \     data_key = :data_key, \
      \     metadata_key = :metadata_key \
      \ WHERE id = :id \
      \;"
    binds () =
      [ Bind ":salt" salt,
        Bind ":data_key" dataKey,
        Bind ":metadata_key" metadataKey,
        Bind ":id" id
      ]

-- | Finds existing user by their email
findByEmail :: Text -> Database -> IO (Maybe User)
findByEmail email db =
  exec db (Statement sql binds $ rowMaybe userRow) ()
  where
    sql =
      "SELECT \
      \ id, email, salt, data_key, metadata_key, created_at \
      \ FROM users \
      \ WHERE email = :email \
      \;"
    binds () = [Bind ":email" email]

-- | Finds existing user by their ID
findByID :: Int64 -> Database -> IO (Maybe User)
findByID id db =
  exec db (Statement sql binds $ rowMaybe userRow) ()
  where
    sql =
      "SELECT \
      \ id, email, salt, data_key, metadata_key, created_at \
      \ FROM users \
      \ WHERE id = :id \
      \;"
    binds () = [Bind ":id" id]

-- | Finds all users
findAll :: Database -> IO (Vector User)
findAll db =
  exec db (Statement sql (const []) $ rowVector userRow) ()
  where
    sql =
      "SELECT \
      \ id, email, salt, data_key, metadata_key, created_at \
      \ FROM users \
      \;"

userRow :: Row User
userRow =
  User
    <$> (column :: Row Int64)
    <*> (column :: Row Text)
    <*> (column :: Row (Maybe Text))
    <*> (column :: Row (Maybe Text))
    <*> (column :: Row (Maybe Text))
    <*> (column :: Row UTCTime)
