{-# LANGUAGE CPP #-}

module OkVault.Resources
  ( getResource,
    bundleCSS,
    bundleJS,
    listTemplates,
  )
where

import qualified Data.ByteString as BS
import Data.List (filter)
import OkCrypt (genericHashHex)
import OkStd
import System.FilePath (dropExtension, takeDirectory, takeExtension, takeFileName)
import Text.Microstache (PName (..))

#if defined(EMBED_RESOURCES)
import OkVault.Resources.Embedded (resources)
import System.IO.Error (ioError, mkIOError, doesNotExistErrorType)
#else
import Paths_okvault (getDataFileName)
import Data.ByteString (readFile)
import System.Directory (listDirectory, doesDirectoryExist)
import System.FilePath ((</>), makeRelative)
#endif

getResource :: FilePath -> IO ByteString
listResources :: IO [FilePath]
#if defined(EMBED_RESOURCES)
getResource name =
    case lookup name resources of
        Just contents -> pure contents
        Nothing -> ioError $ mkIOError doesNotExistErrorType "resource does not exist" Nothing (Just name)
listResources =
  pure $ fst <$> resources
#endif

#if !defined(EMBED_RESOURCES)
getResource name = do
    path <- getDataFileName $ "resources/" <> name
    readFile path
listResources =
  getDataFileName "resources" >>= \p -> walk p p
  where
    walk root path =
      doesDirectoryExist path >>= \case
        False -> pure [makeRelative root path]
        True -> do
          children <- fmap (path </>) <$> listDirectory path
          concat <$> traverse (walk root) children
#endif

bundleCSS :: IO (Text, ByteString)
bundleCSS =
  bundle
    ".css"
    [ "css/vendor/tachyons_4.12.min.css",
      "css/inter.css",
      "css/okvault.css"
    ]

bundleJS :: IO (Text, ByteString)
bundleJS =
  bundle
    ".js"
    [ "js/vendor/sodium_0.7.9.js",
      "js/shortcuts.js",
      "js/crypt.js",
      "js/vault.js",
      "js/item.js",
      "js/pwgen.js",
      "js/totp.js",
      "js/first_time_page.js",
      "js/master_password_prompt.js",
      "js/items_list_page.js",
      "js/add_edit_item_page.js",
      "js/show_item_page.js",
      "js/delete_item_page.js",
      "js/main.js"
    ]

bundle :: Text -> [FilePath] -> IO (Text, ByteString)
bundle ext files = do
  content <- BS.concat <$> traverse getResource files
  hash <- genericHashHex content
  pure ("okvault-" <> hash <> ext, content)

listTemplates :: IO [(PName, FilePath)]
listTemplates = do
  fmap (\t -> (PName $ toText $ dropExtension $ takeFileName t, t)) . filter isTemplate <$> listResources
  where
    isTemplate r = takeExtension r == ".html" && takeDirectory r == "templates"
