module OkVault.Templates
  ( errorPage,
    loginPage,
    loginCodePage,
    firstTimePage,
    logoutPage,
    itemsList,
    addEditItem,
    showItem,
    deleteItem,
  )
where

import qualified Data.Aeson as Aeson
import Data.Aeson.Types (ToJSON (..), object, (.=))
import qualified Data.ByteString.Lazy as LBS
import Data.Text.Lazy (toStrict)
import OkStd
import OkVault.Context
import OkVault.Models
import qualified OkVault.Models.Item as Item
import qualified OkVault.Models.User as User
import Text.Microstache
  ( PName,
    renderMustache,
    templateActual,
  )

renderTemplate :: PName -> Aeson.Value -> InContext ByteString
renderTemplate pname params =
  encodeUtf8 $ toStrict $ renderMustache (templates {templateActual = pname}) params
  where
    templates = getTemplates ?ctx

errorPage :: Text -> Text -> InContext ByteString
errorPage header description =
  renderTemplate "error_page" $
    object
      [ "title" .= Aeson.String "Error",
        "header" .= header,
        "description" .= description
      ]

loginPage :: InContext ByteString
loginPage =
  renderTemplate "login_page" $
    object ["title" .= Aeson.String "Login"]

loginCodePage :: Text -> InContext ByteString
loginCodePage email =
  renderTemplate "login_code_page" $
    object
      [ "title" .= Aeson.String "Enter Code",
        "email" .= email
      ]

firstTimePage :: Text -> InContext ByteString
firstTimePage csrfToken =
  renderTemplate "first_time" $
    object
      [ "title" .= Aeson.String "Welcome!",
        "csrf_token" .= csrfToken
      ]

logoutPage :: InContext ByteString
logoutPage =
  renderTemplate "logout_page" $
    object ["title" .= Aeson.String "Logout"]

itemsList :: User -> Vector Item -> InContext ByteString
itemsList user items =
  renderTemplate "items_list" $
    object
      [ "title" .= Aeson.String "Items",
        "items" .= if null items then Nothing else Just (jsonToText $ itemToJson <$> items),
        "salt" .= mustGetUserSalt user,
        "metadata_key" .= mustGetUserMetadataKey user
      ]
  where
    itemToJson item =
      object
        [ "slug" .= Item.slug item,
          "name" .= Item.name item,
          "category" .= Item.category item
        ]

addEditItem :: User -> Maybe Item -> Text -> InContext ByteString
addEditItem user maybeItem csrfToken =
  renderTemplate "add_edit_item" $
    object
      [ "title" .= (maybe "New Item" (const "Edit Item") maybeItem :: Text),
        "item" .= jsonToText (itemToJson <$> maybeItem),
        "csrf_token" .= csrfToken,
        "salt" .= mustGetUserSalt user,
        "data_key" .= mustGetUserDataKey user,
        "metadata_key" .= mustGetUserMetadataKey user
      ]
  where
    itemToJson item =
      object
        [ "slug" .= Item.slug item,
          "name" .= Item.name item,
          "category" .= Item.category item,
          "content" .= Item.content item,
          "key" .= Item.itemKey item
        ]

showItem :: User -> Item -> InContext ByteString
showItem user item =
  renderTemplate "show_item" $
    object
      [ "title" .= Aeson.String "Item",
        "item" .= jsonToText itemJson,
        "salt" .= mustGetUserSalt user,
        "data_key" .= mustGetUserDataKey user,
        "metadata_key" .= mustGetUserMetadataKey user
      ]
  where
    itemJson =
      object
        [ "slug" .= Item.slug item,
          "name" .= Item.name item,
          "category" .= Item.category item,
          "content" .= Item.content item,
          "key" .= Item.itemKey item
        ]

deleteItem :: User -> Item -> Text -> InContext ByteString
deleteItem user item csrfToken =
  renderTemplate "delete_item" $
    object
      [ "title" .= Aeson.String "Delete Item",
        "item" .= jsonToText itemJson,
        "salt" .= mustGetUserSalt user,
        "csrf_token" .= csrfToken,
        "metadata_key" .= mustGetUserMetadataKey user
      ]
  where
    itemJson =
      object
        [ "slug" .= Item.slug item,
          "name" .= Item.name item
        ]

jsonToText :: ToJSON a => a -> Text
jsonToText = decodeUtf8 . LBS.toStrict . Aeson.encode

mustGetUserSalt :: User -> Text
mustGetUserSalt = fromMaybe (bug "user does not have salt") . User.salt

mustGetUserDataKey :: User -> Text
mustGetUserDataKey = fromMaybe (bug "user does not have data key") . User.dataKey

mustGetUserMetadataKey :: User -> Text
mustGetUserMetadataKey = fromMaybe (bug "user does not have metadata key") . User.metadataKey