module Web.Request
  ( parseCookies,
    createCookie,
    parseForm,
    remoteHost,
  )
where

import Data.Binary.Builder (toLazyByteString)
import Data.ByteString.Lazy (toStrict)
import Data.Time (UTCTime)
import Network.HTTP.Types (Header)
import Network.Wai (Request)
import qualified Network.Wai as Wai
import Network.Wai.Parse
  ( defaultParseRequestBodyOptions,
    lbsBackEnd,
    parseRequestBodyEx,
    setMaxRequestFileSize,
  )
import OkStd
import Web.Cookie hiding (parseCookies)

-- | Parses cookies from the specified request
parseCookies :: Request -> [(Text, Text)]
parseCookies req =
  parseCookiesText cookies
  where
    headers = Wai.requestHeaders req
    cookies = fromMaybe "" $ lookup "Cookie" headers

-- | @createCookie name value expiresAt isSecure@ creates new 'Set-Cookie'
-- header with the given parameters. Cookie is made HTTP-only and its path is
-- set to '/'.
createCookie :: ByteString -> ByteString -> UTCTime -> Bool -> Header
createCookie name value expiresAt isSecure =
  ("Set-Cookie", toStrict $ toLazyByteString $ renderSetCookie cookie)
  where
    cookie =
      defaultSetCookie
        { setCookieName = name,
          setCookieValue = value,
          setCookiePath = Just "/",
          setCookieExpires = Just expiresAt,
          setCookieHttpOnly = True,
          setCookieSecure = isSecure,
          setCookieSameSite = Just sameSiteStrict
        }

-- | Parses form parameters from request body.
parseForm :: Request -> IO [(Text, Text)]
parseForm req = do
  let opts = setMaxRequestFileSize 0 defaultParseRequestBodyOptions
  fmap (bimap decodeUtf8 decodeUtf8) . fst <$> parseRequestBodyEx opts lbsBackEnd req

-- | Returns value of the "X-Forwarded-For" header. If it's absent, returns
-- remote IP and port associated with the request.
remoteHost :: Request -> Text
remoteHost req =
  maybe (toText $ show $ Wai.remoteHost req) decodeUtf8 $ lookup "X-Forwarded-For" $ Wai.requestHeaders req
