class Crypt {
  static async init() {
    return new Promise((resolve, _) => {
      if ('sodium' in window) {
        window.sodium.ready.then(_ => resolve(new Crypt(window.sodium)));
        return;
      }

      window.sodium = {
        onload: function(sodium) {
          resolve(new Crypt(sodium));
        }
      };
    });
  };

  get ciphertextHeader() {
    return this.sodium.from_string('encdat01');
  }

  constructor(sodium) {
    this.sodium = sodium;
  }

  deriveKey(password, salt) {
    return this.sodium.crypto_pwhash(
      this.sodium.crypto_secretbox_KEYBYTES,
      password, salt,
      this.sodium.crypto_pwhash_OPSLIMIT_INTERACTIVE,
      this.sodium.crypto_pwhash_MEMLIMIT_INTERACTIVE,
      this.sodium.crypto_pwhash_ALG_ARGON2ID13);
  }

  generateSalt() {
    return this.sodium.randombytes_buf(this.sodium.crypto_pwhash_SALTBYTES);
  }

  generateKey() {
    return this.sodium.randombytes_buf(this.sodium.crypto_secretbox_KEYBYTES);
  }

  generateNonce() {
    return this.sodium.randombytes_buf(this.sodium.crypto_secretbox_NONCEBYTES);
  }

  encrypt(plaintext, key) {
    const nonce = this.generateNonce();
    const ciphertext = this.sodium.crypto_secretbox_easy(plaintext, nonce, key);
    return this.toBase64(new Uint8Array([
      ...this.ciphertextHeader,
      ...nonce,
      ...ciphertext
    ]));
  }

  decrypt(ciphertext, key) {
    const header = this.ciphertextHeader;
    const cipherdata = this.fromBase64(ciphertext);
    if (!cipherdata.subarray(0, header.length).every((e, i) => e === header[i])) {
      throw new Error('Invalid ciphertext header');
    }
    const nonce = cipherdata.subarray(header.length, header.length + this.sodium.crypto_secretbox_NONCEBYTES);
    return this.sodium.crypto_secretbox_open_easy(
      cipherdata.subarray(header.length + this.sodium.crypto_secretbox_NONCEBYTES),
      nonce,
      key);
  }

  decryptString(ciphertext, key) {
    return this.toString(this.decrypt(ciphertext, key));
  }

  toBase64(data) {
    return this.sodium.to_base64(data);
  }

  fromBase64(str) {
    return this.sodium.from_base64(str);
  }

  toString(typedArray) {
    return this.sodium.to_string(typedArray);
  }
}
