module TestAuth where

import Data.Maybe (isJust)
import Data.Text (pack, unpack)
import Data.Time.Format.ISO8601 (iso8601ParseM)
import qualified Network.Wai as Wai
import OkSqlite
import OkStd
import qualified OkVault.Actions.Shared as Actions
import qualified OkVault.Context as Context
import OkVault.DatabaseSchema (ensureDatabaseSchema)
import qualified OkVault.Models.Session as Session
import qualified OkVault.Models.User as User
import System.Directory (getTemporaryDirectory, removeFile)
import System.Environment (setEnv)
import System.IO (hClose, openTempFile)
import Test.HUnit

tests =
  [ "should mark request as 'not authenticated' when session is not found" ~: withTempDB $ \_ -> do
      ctx <- Context.fromDatabase
      let ?ctx = ctx
      let ?req = Wai.defaultRequest {Wai.requestHeaders = [("Cookie", "okvault.sid=whatever")]}
      curUser <- Actions.currentUser
      assertEqual "authentication status" False (isJust curUser),

    "should mark request as 'not authenticated' when session is expired" ~: withTempDB $ \dbPath -> do
      sid <- withDatabase dbPath OpenReadWrite $ \db -> do
        User.create "user@test.fake" db
        Just u <- User.findByEmail "user@test.fake" db
        Session.id <$> Session.create (User.id u) (parseISO8601 "1970-01-01T00:00:00Z") db
      ctx <- Context.fromDatabase
      let ?ctx = ctx
      let ?req = Wai.defaultRequest {Wai.requestHeaders = [("Cookie", "okvault.sid=" <> encodeUtf8 sid)]}
      curUser <- Actions.currentUser
      assertEqual "authentication status" False (isJust curUser),

    "should mark request as 'authenticated' when session is not expired" ~: withTempDB $ \dbPath -> do
      (u, sid) <- withDatabase dbPath OpenReadWrite $ \db -> do
        User.create "user@test.fake" db
        Just u <- User.findByEmail "user@test.fake" db
        sid <- Session.id <$> Session.create (User.id u) (parseISO8601 "2100-01-01T00:00:00Z") db
        pure (u, sid)
      ctx <- Context.fromDatabase
      let ?ctx = ctx
      let ?req = Wai.defaultRequest {Wai.requestHeaders = [("Cookie", "okvault.sid=" <> encodeUtf8 sid)]}
      currentUser <- Actions.currentUser
      assertEqual "current user" (Just u) currentUser
  ]

withTempDB =
  bracket
    ( do
        tmpdir <- getTemporaryDirectory
        (dbPathStr, h) <- openTempFile tmpdir "okvault.db"
        hClose h
        let dbPath = pack dbPathStr
        withDatabase dbPath OpenReadWriteCreate ensureDatabaseSchema
        setEnv "DBPATH" dbPathStr
        pure dbPath
    )
    (removeFile . unpack)

parseISO8601 s =
  case iso8601ParseM s of
    Nothing -> error $ "can't parse ISO8601 timestamp from: " <> toText s
    Just t -> t
